// LifeGame.cpp: CLifeGame NX̃Cve[V
//
//////////////////////////////////////////////////////////////////////
//#define STAGE1_ASM 1
#define STAGE2_ASM 1

#include "stdafx.h"
#include <algorithm>
#include "Life.h"
#include "LifeGame.h"
#include "ManagedArray.h"
using namespace std;

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define BIT(x)      ( 1UL << (x) )
#define GEN_MASK(x) ( BIT(x) - 1 )
#define STRIDE(w) \
  ( ((w) >> (BITS_PER_ELEM_LOG - BITS_PER_CELL_LOG)) + 1 )

//////////////////////////////////////////////////////////////////////
// \z/
//////////////////////////////////////////////////////////////////////

CLifeGame::CLifeGame(unsigned long width, unsigned long height)
{
  const unsigned long uSize = STRIDE(width) * height;

  m_width = width, m_height = height;
  m_pMap  = m_pBuffer[0] = new CLifeElem[uSize];
  m_pTemp = m_pBuffer[1] = new CLifeElem[uSize];
  memset(m_pBuffer[0], 0, sizeof CLifeElem * uSize);
  memset(m_pBuffer[1], 0, sizeof CLifeElem * uSize);
}

CLifeGame::~CLifeGame()
{
  if(m_pBuffer[0]) { delete[] m_pBuffer[0]; m_pBuffer[0] = NULL; }
  if(m_pBuffer[1]) { delete[] m_pBuffer[1]; m_pBuffer[1] = NULL; }
}

//////////////////////////////////////////////////////////////////////

static inline
void Generate(CLifeElem *pNextMap, const CLifeElem *pMap,
	      unsigned long width, unsigned long height)
{
  const unsigned long uStride = STRIDE(width), uSize = uStride * height;
  ASSERT((pMap & 3) == 0 && (pNextMap & 3) == 0);
#if     defined(STAGE1_ASM) || defined(STAGE2_ASM)
  _asm {
    push    ebx
    push    ecx
    push    edx
    push    esi
    push    edi
  }
#endif/*defined(STAGE1_ASM) || defined(STAGE2_ASM)*/
  if(height <= 1)
    memcpy(pNextMap, pMap, sizeof CLifeElem * uSize);
  else {
#ifdef  STAGE1_ASM
    _asm {
      mov  esi, pMap
      mov  edi, pNextMap
      mov  ebx, stride
      shl  ebx, 2
      
      mov  ecx, stride
    count_upper_lp:
      mov  eax, dword ptr [esi]
      add  eax, dword ptr [esi + ebx]
      add  esi, 4
      mov  dword ptr [edi], eax
      add  edi, 4
      dec  ecx
      jnz  count_upper_lp
      
    count_upper_end:
      sub  esi, ebx
      sub  esi, ebx
      mov  ecx, uSize
      sub  ecx, ebx
      jle  count_mid_end
    count_mid_lp:
      mov  eax, dword ptr [esi]
      mov  edx, esi
      add  edx, ebx
      add  eax, dword ptr [edx]
      add  eax, dword ptr [esi + ebx * 2]
      mov  dword ptr [edi], eax
      add  esi, 4
      add  edi, 4
      dec  ecx
      jnz  count_mid_lp
      
    count_mid_end:
      mov  ecx, stride
    count_lower_lp:
      mov  eax, dword ptr [esi]
      add  eax, dword ptr [esi + ebx]
      mov  dword ptr [edi], eax
      add  esi, 4
      add  edi, 4
      dec  ecx
      jnz  count_lower_lp
    count_end:
    }
#else /*! STAGE1_ASM */
    register unsigned long i;
    register CLifeElem *pCursor = pNextMap;
    register const CLifeElem *pOrig = pMap;

    for(i = uStride; i-- > 0; pOrig++)
      *pCursor++ = pOrig[0] + pOrig[uStride];
    for(i = uSize - 2 * uStride; i-- > 0; pOrig++)
      *pCursor++ = pOrig[0] + pOrig[uStride] + pOrig[-uStride];
    for(i = uStride; i-- > 0; pOrig++)
      *pCursor++ = pOrig[0] + pOrig[-uStride];
#endif/*STAGE1_ASM */
  }

#ifdef  STAGE2_ASM
  _asm {
    push ebp		  // 55, 4uop
    mov  esi, pMap	  // 8B 7x xx, 1uop
    mov  edi, pNextMap	  // 8B 7x xx, 1uop
    mov  ecx, uSize	  // 8B 5x xx, 1uop
    dec  ecx		  // 49, 1uop

    even
    ; load
    xor  edx, edx	  // 31 D2, 1uop
    mov  eax, dword ptr [edi] // 8B 07, 1uop
  gen_core_lp: ; EDX : EAX : [EDI + 4]
    ; left, mid [0000h]
    mov  ebx, eax   // 8B D8, 1uop
    shl  ebx, 4	    // C1 E3 04, 4uop[dec0]
    shr  edx, 28    // C0 EA 1C, 1uop
    add  edx, eax   // 03 D0, 1uop
    add  ebx, edx   // 03 DA, 1uop
    mov  edx, eax   // 8B D0, 1uop
    shr  eax, 4	    // C0 E8 04, 1uop
    add  ebx, eax   // 03 D8, 1uop
    ; right [0013h]
    mov  eax, dword ptr [edi + 4] // 8B 47 04, 1uop
    mov  ebp, eax   // 8B E8, 1uop
    shl  eax, 28    // C1 E0 1C, 4uop[dec0]
    add  eax, ebx   // 03 C3, 1uop
    
    ; postprocess [001Dh]
    mov  ebx, dword ptr [esi] // 8B 1E, 1uop
    add  esi, 4	    // 83 C6 04, 1uop
    sub  eax, ebx   // 2B C3, 1uop
    or   eax, ebx   // 09 D8, 1uop
    xor  eax, 0CCCCCCCCh  // 35 CC CC CC CC, 1uop
    mov  ebx, eax   // 8B D8, 1uop
    shr  ebx, 2	    // C1 EB 02, 1uop
    and  eax, ebx   // 23 C3, 1uop
    mov  ebx, eax   // 8B D8, 1uop
    shr  ebx, 1	    // D1 EB, 1uop
    and  eax, ebx   // 23 C3, 1uop
    and  eax, 11111111h	  // 25 11 11 11 11, 1uop
    mov  dword ptr [edi], eax // 89 07, 2uop
    add  edi, 4	    // 83 C7 04, 1uop
    
    ; [0042h]
    mov  eax, ebp   // 8B C5, 1uop
    dec  ecx	    // 49, 1uop
    jnz  gen_core_lp	  // 75 xx, 1uop
    ; [0047h]
    
  gen_last:
    ; left, mid
    mov  ebx, eax
    shl  ebx, 4
    shr  edx, 28
    add  edx, eax
    add  ebx, edx
    mov  edx, eax
    shr  edx, 4
    add  eax, ebx
    add  eax, edx
    
    ; postprocess
    mov  ebx, dword ptr [esi]
    sub  eax, ebx
    or   eax, ebx
    xor  eax, 0CCCCCCCCh
    mov  ebx, eax
    shr  ebx, 2
    and  eax, ebx
    mov  ebx, eax
    shr  ebx, 1
    and  eax, ebx
    and  eax, 11111111h
    mov  dword ptr [edi], eax
    
    ; last mask
    pop  ebp

    ; generate mask
    mov  ecx, width	// 1uop
    and  ecx, 7		// 1uop
    shl  ecx, 2		// 4uop
    mov  eax, 1		// 1uop
    mov	 ebx, uStride	// 1uop
    shl  eax, cl	// 4uop
    dec  eax		// 1uop
    mov	 ecx, height	// 1uop
    shl  ebx, 2		// 4uop
    shr  ecx, 1		// 1uop
    jnc  mask_lp_entry	// 1uop
    and  dword ptr [edi], eax // 4uop
    or   ecx, ecx	// 1uop
  mask_lp_entry:
    jz   mask_end	// 1uop
#if 0
  mask_lp:
    sub  edi, ebx	// 1uop
    sub	 edi, ebx	// 1uop
    and  dword ptr [edi], eax // 4uop
    and  dword ptr [edi + ebx], eax // 4uop
    dec  ecx		// 1uop
    jnz  mask_lp	// 1uop
#else
    sub  edi, ebx	// 1uop
  mask_lp:
    and  dword ptr [edi], eax // 4uop
    sub  edi, ebx	// 1uop
    sub	 edi, ebx	// 1uop
    and  dword ptr [edi + ebx], eax // 4uop
    dec  ecx		// 1uop
    jnz  mask_lp	// 1uop
#endif
  mask_end:
  }
#else /*! STAGE2_ASM*/
  {
    register CLifeElem *pCursor = pNextMap;
    register const CLifeElem *pOrig = pMap;
    register CLifeElem prev = 0UL, temp;
    register unsigned long i;

    for(i = uSize; --i > 0; pOrig++) {
      temp = pCursor[0] + (pCursor[0] << BITS_PER_CELL) + (pCursor[0] >> BITS_PER_CELL);
      temp += prev >> BITS_PER_ELEM - BITS_PER_CELL;
      temp += pCursor[1] << BITS_PER_ELEM - BITS_PER_CELL;
      
      prev = pCursor[0];
      temp = temp - *pOrig | *pOrig;
      temp ^= 0xCCCCCCCC;
      temp &= temp >> 2;
      temp &= temp >> 1;
      *pCursor++ = temp & 0x11111111;
    }
    temp = pCursor[0] + (pCursor[0] << BITS_PER_CELL) + (pCursor[0] >> BITS_PER_CELL);
    temp += prev >> BITS_PER_ELEM - BITS_PER_CELL;
  
    prev = pCursor[0];
    temp = temp - *pOrig | *pOrig;
    temp ^= 0xCCCCCCCC;
    temp &= temp >> 2;
    temp &= temp >> 1;
    *pCursor = temp & 0x11111111;
  
    temp = GEN_MASK(width & GEN_MASK(BITS_PER_ELEM_LOG - BITS_PER_CELL_LOG));
    while(height--) {
      *pCursor &= temp;
      pCursor -= uStride;
    }
  }
#endif/*STAGE2_ASM*/
#if     defined(STAGE1_ASM) || defined(STAGE2_ASM)
  _asm {
    pop	    edi
    pop	    esi
    pop	    edx
    pop	    ecx
    pop	    ebx
  }
#endif/*defined(STAGE1_ASM) || defined(STAGE2_ASM)*/
}

static inline
void GenerateBitmap(BYTE *pBitmap, const CLifeElem *pMap,
		    unsigned long width, unsigned long height)
{
  const unsigned long delta_elem =
    sizeof CLifeElem *
    (STRIDE(width) -
     ((width + GEN_MASK(BITS_PER_ELEM_LOG - BITS_PER_CELL_LOG))
      >> (BITS_PER_ELEM_LOG - BITS_PER_CELL_LOG))
     );
  const unsigned long delta_bitmap = ((width + 7) >> 3) & 1;

#if 1
  _asm {
    push ebx
    push ecx
    push edx
    push esi
    push edi

    mov  esi, pMap
    mov  edi, pBitmap
    mov  ebx, height
  vert_lp:
    mov  ecx, width
    add  ecx, 7
    shr  ecx, 3
  horz_lp:
    mov  edx, dword ptr [esi]	// 2uop
    xor  al, al	  // 1uop
    add  esi, 4	  // 1uop
    shr  edx, 1	  // 1uop
    adc  al, 0	  // 2uop
    shl  al, 1	  // 4uop
    shr  edx, 4	  // 1uop
    adc  al, 0	  // 2
    shl  al, 1	  // 4
    shr  edx, 4	  // 1
    adc  al, 0
    shl  al, 1
    shr  edx, 4
    adc  al, 0
    shl  al, 1
    shr  edx, 4
    adc  al, 0
    shl  al, 1
    shr  edx, 4
    adc  al, 0
    shl  al, 1
    shr  edx, 4
    adc  al, 0
    shl  al, 1
    shr  edx, 4
    adc  al, 0
    not  al
    mov  byte ptr [edi], al
    inc  edi
    dec  ecx
    jnz  horz_lp
    add  esi, delta_elem
    add  edi, delta_bitmap
    dec  ebx
    jnz  vert_lp

    pop  edi
    pop  esi
    pop  edx
    pop  ecx
    pop  ebx
  }
#else
  while(height--) {
    register unsigned long count = width;
    while(count--) {
      ;
    }
  }
#endif
}

//////////////////////////////////////////////////////////////////////
// C^[tFCX

BOOL CLifeGame::IsCellSet(unsigned long x, unsigned long y) const
{
  ASSERT(x < m_width && y < m_height);

  register CLifeElem elm =
    m_pMap[y * STRIDE(m_width) + (x >> (BITS_PER_ELEM_LOG - BITS_PER_CELL_LOG))];
  elm >>= BITS_PER_CELL * (x & GEN_MASK(BITS_PER_ELEM_LOG - BITS_PER_CELL_LOG));
  return elm & GEN_MASK(BITS_PER_ELEM);
}

void CLifeGame::SetCell(unsigned long x, unsigned long y, BOOL set)
{
  ASSERT(x < m_width && y < m_height);

  const register unsigned long idx =
    y * STRIDE(m_width) + (x >> (BITS_PER_ELEM_LOG - BITS_PER_CELL_LOG));
  register unsigned long shift =
    BITS_PER_CELL * (x & GEN_MASK(BITS_PER_ELEM_LOG - BITS_PER_CELL_LOG));
  m_pMap[idx] &= ~(GEN_MASK(BITS_PER_CELL) << shift);
  m_pMap[idx] |= (set ? BIT(shift) : 0);
}

void CLifeGame::Next()
{
  Generate(m_pTemp, m_pMap, m_width, m_height);
  swap(m_pMap, m_pTemp);
}

void CLifeGame::MakeBits(void *pBuffer) const
{
  GenerateBitmap((BYTE *) pBuffer, m_pMap, m_width, m_height);
}

void CLifeGame::Resize(unsigned long width, unsigned int height)
{
  const unsigned long uNewSize = STRIDE(width) * height;

  // create new buffer and
  CLifeElem *pNewBuffer = new CLifeElem[uNewSize];
  memset(pNewBuffer, 0, sizeof CLifeElem * uNewSize);

  // copy
  for(unsigned long y = 0; y < min(m_height, height); y++)
    memcpy(&pNewBuffer[STRIDE(width) * y], &m_pMap[STRIDE(m_width) * y],
	   sizeof CLifeElem * STRIDE(min(width, m_width)));

  // replace bitmap
  m_pMap  = m_pBuffer[0] = pNewBuffer;
  // and prepare temporary buffer
  m_pTemp = m_pBuffer[1] = pNewBuffer = new CLifeElem[uNewSize];
  memset(pNewBuffer, 0, sizeof CLifeElem * uNewSize);
  m_width = width, m_height = height;
}
